# This is a practice program for people learning to use Numpy and Scipy.
# Copyright Sam Kennerly 2010 (Creative Commons Attribution/Noncommercial/Share-Alike 3.0)
# http://creativecommons.org/licenses/by-nc-sa/3.0/

# TO RUN THIS PROGRAM:
# Open a terminal and change to the directory containing this file, then type "python eigensystem.py" .

# This program does some popular matrix tricks, including finding eigenvalues and eigenvectors.
# You can type in a square matrix by hand or generate a random square matrix.
# Eigenvalues are plotted in the complex plane along with histograms of their polar coordinates.

from numpy import *		# general NumPy stuff
from scipy import *		# general SciPy stuff
from pylab import *		# general PyLab stuff
import math				# general math stuff
import random			# random number features from Python
	
mat_size = int(raw_input("What size matrix? "))
# raw_input() always produces a string.  int() forces it to be an integer.

M = zeros((mat_size,mat_size))
# This creates a square Numpy array of size mat_size and fills it with zeros.
M = mat(M)
# This makes M into an official Numpy matrix.

manual = raw_input("Press M to input a matrix manually, or Enter for a random matrix. ").upper()
# upper() forces this character to be upper-case so typing 'm' and 'M' are both OK.

skip_vectors = raw_input("Press S to skip finding eigenvectors.  (This saves some time.) ").upper()
# Do we want to find eigenvectors and eigenvalues, or just eigenvalues?

# This part lets the user type in rows of the matrix, if he/she wants to.
if (manual == "M") :	# Need the double-equal-sign because we're testing a true/false statement.
	print "Type the matrix elements. Separate elements by spaces, hit Enter after each row.\n"
	for i in range(mat_size) :
		M[i,:] = raw_input().split()	# split() chops the input into individual strings.
		# Matrix elements are labeled like this: M[3,4] is the element in row 3, column 4.  
		# The : is a "wildcard" character.  M[i,:] means "ith row, any column of M."

# If the user didn't type M (or m), then we'll make a random matrix.
else :
	for i in range(mat_size) :
		for j in range(mat_size) :
			M[i,j] = random.uniform(-1,1)
			# Each element is a random number chosen uniformly from -1.0 to 1.0 .
			# If you don't want a random matrix, type your own formula here!

# Print M to screen so you can see what a NumPy matrix looks like.
print "\n The matrix you typed is:\n", M

# Here are some useful matrix tricks that NumPy can do automatically:		
trace = M.trace()
determinant = linalg.det(M)
Msquared = dot(M,M)
print "\nTrace:", trace, "\tDeterminant:", determinant, "\n\nMatrix squared:\n", Msquared
# \n means 'new line' and \t means 'tab character.'

# Find the inverse of M, but be careful: if det(M) == 0, then M has no inverse.
# If det(M) is close to zero, finding the inverse can cause numerical problems.
if ( abs(determinant) < 10**(-14) ) :
	print "\nWarning: Singular or near singular matrix!  Cannot safely be inverted.\n"
else :
	print "\nInverse:\n", linalg.inv(M), "\n"

# Here's yet another matrix trick.  Non-symmetric matrices can have complex eigenvalues.
# If M is symmetric, it will equal its own transpose.  That means we can use a somewhat
# faster algorithm later when we find eigenvalues and eigenvectors.
M_is_symmetric = True			# This is a boolean variable: it's either True or False.
SymCheck = M - M.transpose()	# SymCheck is a matrix.  If M is symmetric, it's all zeros.
if ( SymCheck.any() != 0 ) :	# Are any elements of SymCheck NOT equal to zero?
	M_is_symmetric = False
	print "Warning: matrix is not symmetric. Eigenvalues and eigenvectors may be complex."

# Now we'll find the eigenvalues using NumPy functions.
if (skip_vectors == "S") :
	evalues = linalg.eigvals(M)
	# evalues is a 1 by mat_size array of eigenvalues.

# If we want eigenvectors, we should use eigh (for symmetric matrices) or eig (for any square matrix).
# eig and eigh return two arrays: evalues is a 1 by mat_size array of eigenvalues, and
# evectors is a square array in which each column is an eigenvector.
elif (M_is_symmetric == True) :			# elif means 'else if'
	evalues, evectors = linalg.eigh(M)
else :
	evalues, evectors = linalg.eig(M)

# Sorting eigenvalues and eigenvectors can be a bit confusing.
# The function argsort() lets us write them in order from smallest to largest.
# If the eigenvalues are complex, it ranks them by real part.
rankings = argsort(evalues)
evalues = evalues[rankings]

# If we found eigenvectors, we should sort them as well.
if (skip_vectors != "S") :
	evectors = evectors[:,rankings]
	# The eigenvectors are stored as columns of the array called 'evectors.'
	for j in range(mat_size) :
		print "\nEigenvector", evectors[:,j], "\nhas eigenvalue", evalues[j]

# If we skipped finding eigenvectors, we can just print eigenvalues.
else :
	print "\nEigenvalues:\n", evalues, "\n"

# Turn on Pylab interactive mode so the plot window will be visible while the program runs.
ion()

# Let's plot the eigenvalues of M on the complex plane.

figure(1)						# Stuff below here will be in the second picture generated.
subplot(1,2,1)					# Fig. 2 contains 1 row, 2 columns of subplots.  This is #1.
magnitudes = (abs(evalues)**2)	# Make an array of square magnitudes of the eigenvalues.
hist(magnitudes, bins=51)		# Make a histogram plot of magnitudes with 51 bins.
title("square magnitude histogram")
xlabel("square magnitude")
ylabel("number of evalues")
grid(True)
subplot(1,2,2)					# Fig. 2 contains 1 row, 2 columns of subplots.  This is #2.
arguments = angle(evalues)		# Make an array of polar angles of the eigenvalues.
hist(arguments, bins=51)		# Another histogram plot with 51 bins.
title("angle histogram")
xlabel("angle")
ylabel("number of evalues")
grid(True)

figure(2)						# This will be one of two pictures generated.
real_parts = real(evalues)		# Make an array of the real part of the eigenvalues.
imag_parts = imag(evalues)		# Make an array of the imaginary part of the eigenvalues.
scatter(real_parts, imag_parts, s=1)	# Make a scatterplot with real_parts on the x-axis
# and imag_parts on the y-axis.  s=1 means "draw dots of size 1 point."
scatter(real_parts, imag_parts)
# Bug alert! My version of PyLab requires scatter() to be called TWICE or it doesn't work.
title("eigenvalues")
xlabel("real part")
ylabel("imaginary part")
axes().set_aspect('equal')		# Means "don't stretch x- and y-axes by a different amount."
grid(True)						# Make this plot look like graph paper.
savefig('Eigenvalues.png')		# Savefig lets us produce a graphics file.

# These plots disappear as soon as the program closes.  Here's a trick to keep them open:
press_any_key = raw_input("\nPress Enter to quit.")
ioff()		# Turn off interactive mode.