# This is a practice program for people learning to use Numpy and Scipy.
# Copyright Sam Kennerly 2010 (Creative Commons Attribution/Noncommercial/Share-Alike 3.0)
# http://creativecommons.org/licenses/by-nc-sa/3.0/

# TO RUN THIS PROGRAM:
# Open a terminal and change to the directory containing this file, then type "python Demo.py" .

# Python basics: Lines starting with the # symbol are comments.
# In Python, whitespace matters!  Start new lines by hitting Enter.
# You MUST indent lines if you want them to be part of if statements, for/while loops, and other such things.
# Four spaces is the usual Python indent.

from numpy import *		# general NumPy stuff
from scipy import *		# general SciPy stuff
from pylab import *		# general PyLab stuff
from scipy.integrate import *	# stuff for numerical integration
import math				# general math stuff
			
# Now we'll integrate some ordinary differential equations (i.e. one independent variable).
# Partial differential equations are much harder!  We don't do them in this program.
# For all of these examples, x is a single number that depends on time.
# u will be a numpy array of x and its derivatives: u[0]=x, u[1]=x', etc.
# We'll first make some functions that return the derivatives of both x and x' .

# Damped harmonic oscillator x'' + bx' + kx = A cos[wt]
def d_harmonic(u,t,b,k,w,A) :
	xdot = u[1]
	vdot = -b*u[1] - k*u[0] + A*cos(w*t)
	return xdot, vdot
# van der Pol oscillator x'' + b(x^2 -1)x' + kx = A cos[wt]
def d_vanderpol(u,t,b,k,w,A) :
	xdot = u[1]
	vdot = -b*(u[0]*u[0]-1)*u[1] - k*u[0] + A*cos(w*t)
	return xdot, vdot
# Duffing oscillator x'' + bx' + kx(1+cx^2) = A cos[wt]
# (This example is actually just the special case c=0.1.)
def d_duffing(u,t,b,k,w,A) :
	xdot = u[1]
	vdot = -b*u[1] - k*u[0]*(1+0.1*u[0]*u[0]) + A*cos(w*t)
	return xdot, vdot
		

# How many time steps do we want, and over what time period?
endtime, numsteps = double(raw_input("Enter end time and number of timesteps: ").split())
# double() forces these to be double-precision numbers.  But numsteps should really be an integer:
numsteps = int(numsteps)

# t_array will be an array of all the times at which the equation will be estimated.
# linspace() creates an array of evenly-spaced numbers between 0 and endtime.
t_array = linspace(0, endtime, numsteps)

# We need to specify initial values of x and x' if we want numerical solutions.
x0, v0 = double(raw_input("Enter initial values for position and velocity: ").split())
u_initial = array([x0,v0])

# We'll ask the user for values of b, k, w, and A
b, k = double(raw_input("Enter damping and spring constant: ").split())
w, A = double(raw_input("Enter driving frequency (radians/sec) and amplitude: ").split())


# Now for the actual work using numpy's odeint() function:
harmonic = odeint(d_harmonic, u_initial, t_array, args=(b,k,w,A))
vanderpol = odeint(d_vanderpol, u_initial, t_array, args=(b,k,w,A))
duffing = odeint(d_duffing, u_initial, t_array, args=(b,k,w,A))
# I've used args=() to include the user's chosen values of b,k,w,A as "arguments."


# It might be useful to save the results to a text file.
# The 0th column will be time.  The 1st, 2nd and 3rd will be position for
# harmonic, van der Pol, and Duffing equations.  Columns 4,5,6 will be velocities.

# Here's how to create a new file.  We'll open it in "write" mode, so be careful!
# If oscillators.dat already exists, this program will overwrite it.
osc_datafile = open("oscillators.dat", "w")

# Now let's output all this data into the file as columns separated by spaces.
for i in range(numsteps) :
	print >> osc_datafile, t_array[i], harmonic[i,0], vanderpol[i,0], duffing[i,0],
	print >> osc_datafile, harmonic[i,1], vanderpol[i,1], duffing[i,1]
print

# It's a good habit to close files when you're done writing to them.
osc_datafile.close()
print "Output saved to oscillators.dat."

# Turn on Pylab interactive mode so the plot window will be visible while the program runs.
ion()

# Here's a plot of x(t) for all three equations.
figure(1)
plot(t_array,harmonic[:,0])		# Output of odeint is an Nx2 array where N = numsteps
plot(t_array,vanderpol[:,0])	# These lines plot the 0 column, which are x values
plot(t_array,duffing[:,0])		# This will produce plots of position vs. time.
xlabel('time')
ylabel('position')
legend(('harmonic', 'van der Pol', 'Duffing'))
title('Oscillator position')
savefig('Osc_Positions.png')	# Savefig lets us produce a graphics file.

# Here's a plot of x'(t) for all three equations.
figure(2)
plot(t_array,harmonic[:,1])		# These lines plot the 1 column, which are x' values
plot(t_array,vanderpol[:,1])	# Now we're plotting velocity vs. time.
plot(t_array,duffing[:,1])		# t-array is in whatever units we chose for endtime.
xlabel('time')
ylabel('velocity')
legend(('harmonic', 'van der Pol', 'Duffing'))
title('Oscillator velocity')
savefig('Osc_Velocities.png')
print "Plots saved to Osc_Positions.png and Osc_Velocities.png."

# These plots disappear as soon as the program closes.  Here's a trick to keep them open:
press_any_key = raw_input("\nPress Enter to quit.")
ioff()		# Turn off interactive mode.