# This is a practice program for people learning to use Numpy and Scipy.
# Copyright Sam Kennerly 2010 (Creative Commons Attribution/Noncommercial/Share-Alike 3.0)
# http://creativecommons.org/licenses/by-nc-sa/3.0/

# TO RUN THIS PROGRAM:
# Open a terminal and change to the directory containing this file, then type "python roll_dice.py" .

# This program simulates rolling dice, counts how many times each number from 1 to 6 occurs,
# and performs a statistical test to check if each number comes up as often as it should.
# It can also generate random floating-point numbers from 1.0 to 6.999... then count how
# often the leading digit is a number from 1 to 6 and do the same statistical test.

from numpy import *		# general NumPy stuff
from scipy import *		# general SciPy stuff
from pylab import *		# general PyLab stuff
import math				# general math stuff
import random			# random number features from Python
import scipy.stats		# SciPy statistics stuff

print						# Print a blank line.
dicelist = []				# This is an empty list.  We'll fill it up soon.
number_of_dice = int(raw_input("How many dice? "))
# raw_input always creates strings.  I used int() to convert the result into an integer.

# Do we want integers (like real dice) or floating-point numbers?
die_type = raw_input("Press F for floating-point numbers or Enter for integers. ").upper()
# .upper() forces die_type to be uppercase so it's OK to type 'F' or 'f'.

# We'll keep track of the sum of all dice rolled while we're rolling them.
dice_sum = 0.0 
# Initializing dice_sum as 0.0 makes it a float, not an integer.
# Accidentally mixing floats and integers is a common programing disaster.
# For example, 22/10 = 2 using "integer division."  22.0/10.0 = 2.2 is "float division."

# Roll some dice! This loop goes from j=0 to j=(number_of_dice - 1).
for j in range(number_of_dice) :  
	if(die_type == "F") :
		dicelist.append(random.uniform(1,7))
		# Append random floating-point numbers from 1.0 to 6.9999...
	else :
		dicelist.append(random.randrange(1,7,1))
		# Append random ints from 1 through 6 in steps of 1.  That '7' is not a typo!
	dice_sum += dicelist[j]
	print dicelist[j],			# The comma means "Don't start a new line after each print."
print

# Now we use Numpy's histogram function to count how many times each number was rolled.
# Two things are returned: an array containing the number of results in each bin, and
# another array describing the edges of each bin.
bins, bin_edges = histogram(dicelist, bins = 6, range = (1,7), normed = False)
# There are 6 bins. The first contains results from 1 to 1.999..., then 2 to 2.999..., etc.
# (To be precise, the last bin would also include any rolls of 7.0 if that happened.)
# normed = False means "Don't bother to normalize my results as a probability distribution."

# Here's a crude statistical test: average all the values.
# We expect a number near 3.5 for integer dice and 4.0 for floating-point dice.
dice_average = dice_sum / number_of_dice
print "\nAverage roll is", dice_average

# Of course, getting exactly 3.5 (for dice) or 4.0 (for floats) is rare.
# Let's do an exact 2-tailed binomial test on the contents of each bin.
# Null hypothesis is "each outcome happens with probability 1.0/6.0"
test_results = [0,0,0,0,0,0]	# Store the results in a 1x6 array.
print "Binomial test results:"
for k in range(6) :
	test_results[k] = scipy.stats.binom_test(bins[k],number_of_dice, 1.0/6.0)
	print test_results[k],
	# Each of these results is a "p-value" for that particular bin.
	# p-value is "probability of fair dice producing a result at least as weird as what we found."
	# A "weirder" result is one in which abs(expected - observed) > (what we found).

# I'd also like to calculate an "overall weirdness" to test the fairness of my dice.
# I'm using Fisher's meta-analysis method... look it up!
fisher = 0.0
for k in range(5) :
	fisher += -2.0*log(test_results[k])
# Fisher's method is for independent variables.  We have 6 results, but only 5 are independent.
# Out of 1000 rolls, say there were 801 from 1 to 5.  That means there must have been 199 6's.
# For this reason, I'm ignoring the last bin.  (It's arbitrary which one gets ignored.)

print "\nReduced Fisher statistic:", fisher / 10.0, "(0 is perfect, 1 is typical, >2 is bad.)"
# fisher_reduced behaves like a reduced chi-squared test statistic with 2*5=10 degrees of freedom.
# 0 means every number came up exactly 1/6 of the time, 1 is the most likely result for fair dice.
# Big values of this statistic are not often produced by fair dice.

meta_p = scipy.stats.chisqprob(fisher,10)
# This is the p-value for Fisher meta-statistic.  It's my overall-weirdness measure.
# stats.chisqprob(x,d) finds the probability of a random variable being x or greater,
# assuming the random variable follows a chi-square distribution with d degrees of freedom.

# If this p-value is very low, it means that fair dice rarely produce such weird results.
# Frequentist statisticians would say "We reject the null hypothesis with confidence (1-p)."
confidence = 1.0 - meta_p
percent_confidence = 100.0*round(confidence,4)	# Round to 4 digits, display as percent.

# What's a "very low" value of p?  High standards of confidence mean we're less likely to
# detect unfair dice, but also less likely to wrongly accuse innocent dice.
# Remember: 1 time out of 100, perfect dice will produce a p value of 1% or less!

# I've decided 99.5% (p=0.005) is my standard of confidence.
if (confidence > 0.995) :
	print "We REJECT the fair-dice hypothesis with", percent_confidence, "% confidence."
	fails = fails + 1
	
# Another popular standard is 95%.  I'll call those results "suspicious."
elif (0.995 >= confidence > 0.95) :			# elif means 'else if'.
	print "We SUSPECT the fair-dice hypothesis is false with", percent_confidence, "% confidence."

# If p is greater than 5%, then I'll find the dice "not guilty."
else :
	print "We FAIL TO REJECT the fair-dice hypothesis.  Confidence is", percent_confidence, "%."
print

# Turn on Pylab interactive mode so the plot window will be visible while the program runs.
ion()

# Let's plot the results.
figure(1)
subplots_adjust(wspace=0.42)	# This adjusts the extra width between plots.
subplot(1,2,1)					# Plot 1 row, 2 columns of plots.  This is plot #1.
title('results')
xlabel('outcome')
ylabel('# rolls with this outcome')
outcomes = [1,2,3,4,5,6]				# An array containing our desired x-values.
bar(outcomes, bins, label='actual')		# bar() makes a bar graph.
# I'd like to compare our results to their expectation values.
expected_bins = number_of_dice / 6.0 * ones(6)
# ones(6) creates an array [1,1,1,1,1,1], then we multiply those 1's by number_of_dice/6.0 .
bar(outcomes, expected_bins, color='g', alpha=0.42, label='expected')
# Bar graph, green color, alpha between 0 and 1 makes the bars partly transparent.
legend(loc=8)
# legend() makes a legend using graph labels, loc=8 means "location: bottom center"
grid(True)								# Draw a dotted-line grid in the background

# Now we'll plot the p-values, but I want the y-axis to be log(p) instead of just p.
subplot(1,2,2)
title('p-values (log scale)')
xlabel('outcome')
ylabel('small p is more significant')
bar(outcomes, test_results, alpha=0.42, log=True)	# log=True makes a logarithmic y-axis.
axis([1,7,.0005, 1])	# The x-axis should go from 1 to 7, the y-axis from .0005 to 1.
axhline(y=0.5, color='g', label="p = 50%")		# This makes a horizontal axis line at y=0.5.
axhline(y=0.05, color='y', label="p = 5%")		# 'g' is green, 'y' yellow, 'r' red.
axhline(y=0.005, color='r', label="p = 0.5%")
legend()
savefig('Dice.png')								# Savefig lets us produce a graphics file.

# These plots disappear as soon as the program closes.  Here's a trick to keep them open:
press_any_key = raw_input("Press Enter to quit.")
ioff()		# Turn off interactive mode.